@@ -29,6 +29,7 @@
 extern uint32_t CLZ(uint32_t a);
 
 bool GPUindexing;
+bool GPUcompress;
 
 namespace gl {
 #ifdef HAVE_PTHREAD
@@ -84,16 +85,24 @@ GLint countShader;
 GLint transparentShader;
 GLint blendShader;
 GLint zeroShader;
-GLint preSumShader;
-GLint partialSumShader;
+GLint compressShader;
+GLint sum1Shader;
+GLint sum2Shader;
+GLint sum2fastShader;
+GLint sum3Shader;
+
+GLuint fragments;
 
-GLuint countBuffer;
 GLuint offsetBuffer;
-GLuint sumBuffer;
+GLuint indexBuffer;
+GLuint elementsBuffer;
+GLuint countBuffer;
+GLuint globalSumBuffer;
 GLuint fragmentBuffer;
 GLuint depthBuffer;
 GLuint opaqueBuffer;
 GLuint opaqueDepthBuffer;
+GLuint feedbackBuffer;
 
 bool ssbo;
 bool interlock;
@@ -157,7 +166,15 @@ extern void exitHandler(int);
 
 namespace gl {
 
-GLint processors;
+GLint gs2;
+GLint gs;
+GLint g;
+GLuint processors;
+GLuint localSize;
+GLuint blockSize;
+GLuint groupSize;
+//GLint maxgroups;
+GLuint maxSize;
 
 bool outlinemode=false;
 bool ibl=false;
@@ -198,6 +215,9 @@ const picture* Picture;
 string Format;
 int fullWidth,fullHeight;
 int Width,Height;
+GLuint pixels;
+GLuint elements;
+GLuint lastpixels;
 double oWidth,oHeight;
 int screenWidth,screenHeight;
 int maxTileWidth;
@@ -428,7 +448,6 @@ int window;
 
 using utils::statistics;
 statistics S;
-GLint shaderProg,shaderProgColor;
 
 GLTexture2<float,GL_FLOAT> IBLbrdfTex;
 GLTexture2<float,GL_FLOAT> irradiance;
@@ -523,10 +542,82 @@ void noShaders()
   exit(-1);
 }
 
-// Return ceil(log2(n)) where n is a 32 bit unsigned integer.
-uint32_t ceillog2(uint32_t n)
+void initComputeShaders()
 {
-  return 32-CLZ(n-1);
+  string sum1=locateFile("shaders/sum1.glsl");
+  string sum2=locateFile("shaders/sum2.glsl");
+  string sum2fast=locateFile("shaders/sum2fast.glsl");
+  string sum3=locateFile("shaders/sum3.glsl");
+
+  if(sum1.empty() || sum2.empty() || sum2fast.empty() || sum3.empty())
+    noShaders();
+
+  std::vector<ShaderfileModePair> shaders(1);
+  std::vector<std::string> shaderParams;
+
+  shaders[0]=ShaderfileModePair(sum1.c_str(),GL_COMPUTE_SHADER);
+  ostringstream s,s2;
+  s << "LOCALSIZE " << gl::localSize << "u" << endl;
+  shaderParams.push_back(s.str().c_str());
+  s2 << "BLOCKSIZE " << gl::blockSize << "u" << endl;
+  shaderParams.push_back(s2.str().c_str());
+  GLuint rc=compileAndLinkShader(shaders,shaderParams,true,false,true);
+  if(rc == 0) {
+    GPUindexing=false; // Compute shaders are unavailable.
+    if(settings::verbose > 2)
+      cout << "No compute shader support" << endl;
+  } else {
+//    glGetIntegeri_v(GL_MAX_COMPUTE_WORK_GROUP_COUNT,0,&maxgroups);
+//    maxgroups=min(1024,maxgroups/(GLint) (localSize*localSize));
+    camp::sum1Shader=rc;
+
+    shaders[0]=ShaderfileModePair(sum2.c_str(),GL_COMPUTE_SHADER);
+    camp::sum2Shader=compileAndLinkShader(shaders,shaderParams,true,false,
+                                          true);
+
+    shaders[0]=ShaderfileModePair(sum2fast.c_str(),GL_COMPUTE_SHADER);
+    camp::sum2fastShader=compileAndLinkShader(shaders,shaderParams,true,false,
+                                              true);
+
+    shaders[0]=ShaderfileModePair(sum3.c_str(),GL_COMPUTE_SHADER);
+    camp::sum3Shader=compileAndLinkShader(shaders,shaderParams,true,false,
+                                          true);
+  }
+}
+
+void initBlendShader()
+{
+  string screen=locateFile("shaders/screen.glsl");
+  string blend=locateFile("shaders/blend.glsl");
+
+  if(screen.empty() || blend.empty())
+    noShaders();
+
+  std::vector<ShaderfileModePair> shaders(2);
+  std::vector<std::string> shaderParams;
+
+  ostringstream s;
+  s << "ARRAYSIZE " << maxSize << "u" << endl;
+  shaderParams.push_back(s.str().c_str());
+  if(GPUindexing)
+    shaderParams.push_back("GPUINDEXING");
+  if(GPUcompress)
+    shaderParams.push_back("GPUCOMPRESS");
+  shaders[0]=ShaderfileModePair(screen.c_str(),GL_VERTEX_SHADER);
+  shaders[1]=ShaderfileModePair(blend.c_str(),GL_FRAGMENT_SHADER);
+  camp::blendShader=compileAndLinkShader(shaders,shaderParams,ssbo);
+}
+
+// Return the smallest power of 2 greater than or equal to n.
+inline GLuint ceilpow2(GLuint n)
+{
+  --n;
+  n |= n >> 1;
+  n |= n >> 2;
+  n |= n >> 4;
+  n |= n >> 8;
+  n |= n >> 16;
+  return ++n;
 }
 
 void initShaders()
@@ -534,21 +625,21 @@ void initShaders()
   Nlights=nlights == 0 ? 0 : max(Nlights,nlights);
   Nmaterials=max(Nmaterials,nmaterials);
 
-  shaderProg=glCreateProgram();
-
-  string zero=locateFile("shaders/count0.glsl");
+  string zero=locateFile("shaders/zero.glsl");
+  string compress=locateFile("shaders/compress.glsl");
   string vertex=locateFile("shaders/vertex.glsl");
+  string count=locateFile("shaders/count.glsl");
   string fragment=locateFile("shaders/fragment.glsl");
-  string blend=locateFile("shaders/blend.glsl");
   string screen=locateFile("shaders/screen.glsl");
-  string pre=locateFile("shaders/presum.glsl");
-  string partial=locateFile("shaders/partialsum.glsl");
 
-  if(vertex.empty() || fragment.empty() || blend.empty() || screen.empty() ||
-     zero.empty() || pre.empty() || partial.empty())
+  if(zero.empty() || compress.empty() || vertex.empty() || fragment.empty() ||
+     screen.empty() || count.empty())
     noShaders();
 
-  std::vector<ShaderfileModePair> shaders(1);
+  if(GPUindexing)
+    initComputeShaders();
+
+  std::vector<ShaderfileModePair> shaders(2);
   std::vector<std::string> shaderParams;
 
   if(ibl) {
@@ -556,46 +647,17 @@ void initShaders()
     initIBL();
   }
 
-#ifdef HAVE_SSBO
-  if(GPUindexing) {
-    shaders[0]=ShaderfileModePair(pre.c_str(),GL_COMPUTE_SHADER);
-    GLuint rc=compileAndLinkShader(shaders,shaderParams,true,interlock,true);
-    if(rc == 0) {
-      GPUindexing=false; // Compute shaders are unavailable.
-      if(settings::verbose > 2)
-        cout << "No compute shader support" << endl;
-    } else {
-      camp::preSumShader=rc;
-
-      ostringstream s,S;
-      s << "PROCESSORS " << processors << "u" << endl;
-      shaderParams.push_back(s.str().c_str());
-
-      S << "STEPSM1 " << ceillog2(processors)-1 << "u" << endl;
-      shaderParams.push_back(S.str().c_str());
-      shaders[0]=ShaderfileModePair(partial.c_str(),GL_COMPUTE_SHADER);
-      camp::partialSumShader=compileAndLinkShader(shaders,shaderParams,
-                                                  true,interlock,true);
-    }
-  }
-#endif
-  string count=locateFile(GPUindexing ? "shaders/offset.glsl" :
-                          "shaders/count.glsl");
-  if(count.empty())
-    noShaders();
-
-  shaders.push_back(ShaderfileModePair());
-
   shaders[0]=ShaderfileModePair(vertex.c_str(),GL_VERTEX_SHADER);
 
 #ifdef HAVE_SSBO
+  if(GPUindexing)
+    shaderParams.push_back("GPUINDEXING");
+  if(GPUcompress)
+    shaderParams.push_back("GPUCOMPRESS");
   shaders[1]=ShaderfileModePair(count.c_str(),GL_FRAGMENT_SHADER);
   camp::countShader=compileAndLinkShader(shaders,shaderParams,true);
-  if(camp::countShader) {
+  if(camp::countShader)
     shaderParams.push_back("HAVE_SSBO");
-    if(GPUindexing)
-      shaderParams.push_back("GPUINDEXING");
-  }
 #else
   camp::countShader=0;
 #endif
@@ -630,7 +692,7 @@ void initShaders()
   if(interlock) shaderParams.push_back("HAVE_INTERLOCK");
   camp::materialShader[0]=compileAndLinkShader(shaders,shaderParams,ssbo,
                                                interlock);
-  if(!camp::materialShader[0]) {
+  if(interlock && !camp::materialShader[0]) {
     shaderParams.pop_back();
     interlock=false;
     camp::materialShader[0]=compileAndLinkShader(shaders,shaderParams,ssbo);
@@ -650,7 +712,7 @@ void initShaders()
   shaderParams.pop_back();
 
   shaderParams.push_back("GENERAL");
-  if(Mode == 2)
+  if(Mode != 0)
     shaderParams.push_back("WIREFRAME");
   camp::generalShader[0]=compileAndLinkShader(shaders,shaderParams,ssbo,
                                               interlock);
@@ -662,32 +724,48 @@ void initShaders()
   camp::transparentShader=compileAndLinkShader(shaders,shaderParams,ssbo,
                                                interlock);
   shaderParams.clear();
-  if(ssbo) {
-    shaders[0]=ShaderfileModePair(screen.c_str(),GL_VERTEX_SHADER);
 
+  if(ssbo) {
     if(GPUindexing)
       shaderParams.push_back("GPUINDEXING");
+    shaders[0]=ShaderfileModePair(screen.c_str(),GL_VERTEX_SHADER);
+    shaders[1]=ShaderfileModePair(compress.c_str(),GL_FRAGMENT_SHADER);
+    camp::compressShader=compileAndLinkShader(shaders,shaderParams,ssbo);
+    if(GPUindexing)
+      shaderParams.pop_back();
     else {
       shaders[1]=ShaderfileModePair(zero.c_str(),GL_FRAGMENT_SHADER);
       camp::zeroShader=compileAndLinkShader(shaders,shaderParams,ssbo);
     }
-
-    shaders[1]=ShaderfileModePair(blend.c_str(),GL_FRAGMENT_SHADER);
-    camp::blendShader=compileAndLinkShader(shaders,shaderParams,ssbo);
+    maxSize=1;
+    initBlendShader();
   }
   lastshader=-1;
 }
 
+void deleteComputeShaders()
+{
+  glDeleteProgram(camp::sum1Shader);
+  glDeleteProgram(camp::sum2Shader);
+  glDeleteProgram(camp::sum2fastShader);
+  glDeleteProgram(camp::sum3Shader);
+}
+
+void deleteBlendShader()
+{
+  glDeleteProgram(camp::blendShader);
+}
+
 void deleteShaders()
 {
   if(camp::ssbo) {
-    glDeleteProgram(camp::blendShader);
-    if(GPUindexing) {
-      glDeleteProgram(camp::preSumShader);
-      glDeleteProgram(camp::partialSumShader);
-    } else
+    deleteBlendShader();
+    if(GPUindexing)
+      deleteComputeShaders();
+    else
       glDeleteProgram(camp::zeroShader);
     glDeleteProgram(camp::countShader);
+    glDeleteProgram(camp::compressShader);
   }
 
   glDeleteProgram(camp::transparentShader);
@@ -699,6 +777,13 @@ void deleteShaders()
   glDeleteProgram(camp::pixelShader);
 }
 
+void resizeBlendShader(uint maxsize)
+{
+  gl::maxSize=ceilpow2(maxsize);
+  gl::deleteBlendShader();
+  gl::initBlendShader();
+}
+
 void setBuffers()
 {
   GLuint vao;
@@ -712,10 +797,16 @@ void setBuffers()
   camp::transparentData.Reserve();
 
 #ifdef HAVE_SSBO
-  glGenBuffers(1, &camp::countBuffer);
   glGenBuffers(1, &camp::offsetBuffer);
-  if(GPUindexing)
-    glGenBuffers(1, &camp::sumBuffer);
+  if(GPUindexing) {
+    glGenBuffers(1, &camp::globalSumBuffer);
+    glGenBuffers(1, &camp::feedbackBuffer);
+  }
+  glGenBuffers(1, &camp::countBuffer);
+  if(GPUcompress) {
+    glGenBuffers(1, &camp::indexBuffer);
+    glGenBuffers(1, &camp::elementsBuffer);
+  }
   glGenBuffers(1, &camp::fragmentBuffer);
   glGenBuffers(1, &camp::depthBuffer);
   glGenBuffers(1, &camp::opaqueBuffer);
@@ -803,6 +894,7 @@ void Export()
         trBeginTile(tr);
         remesh=true;
         drawscene(fullWidth,fullHeight);
+        gl::lastshader=-1;
         ++count;
       } while (trEndTile(tr));
       if(settings::verbose > 1)
@@ -955,6 +1047,16 @@ void reshape0(int width, int height)
   Width=width;
   Height=height;
 
+  static int lastWidth=1;
+  static int lastHeight=1;
+  if(View && Width*Height > 1 && (Width != lastWidth || Height != lastHeight)
+     && settings::verbose > 1) {
+    cout << "Rendering " << stripDir(Prefix) << " as "
+         << Width << "x" << Height << " image" << endl;
+    lastWidth=Width;
+    lastHeight=Height;
+  }
+
   setProjection();
   glViewport(0,0,Width,Height);
   if(camp::ssbo)
@@ -1066,7 +1168,7 @@ void togglefitscreen()
 void initTimer()
 {
   gettimeofday(&lasttime,NULL);
-  gettimeofday(&lastframetime,NULL);
+  lastframetime=lasttime;
 }
 
 void idleFunc(void (*f)())
@@ -1102,7 +1204,7 @@ void display()
   bool fps=settings::verbose > 2;
   drawscene(Width,Height);
   if(fps) {
-    if(framecount < 10) // Measure steady-state framerate
+    if(framecount < 20) // Measure steady-state framerate
       seconds();
     else {
       double s=seconds();
@@ -1800,6 +1902,12 @@ void init_osmesa()
 
 #endif /* HAVE_GL */
 
+bool NVIDIA()
+{
+  char *GLSL_VERSION=(char *) glGetString(GL_SHADING_LANGUAGE_VERSION);
+  return string(GLSL_VERSION).find("NVIDIA") != string::npos;
+}
+
 // angle=0 means orthographic.
 void glrender(const string& prefix, const picture *pic, const string& format,
               double width, double height, double angle, double zoom,
@@ -1810,12 +1918,6 @@ void glrender(const string& prefix, const picture *pic, const string& format,
 {
   Iconify=getSetting<bool>("iconify");
 
-#if defined(HAVE_COMPUTE_SHADER) && !defined(HAVE_LIBOSMESA)
-  GPUindexing=getSetting<bool>("GPUindexing");
-#else
-  GPUindexing=false;
-#endif
-
   if(zoom == 0.0) zoom=1.0;
 
   Prefix=prefix;
@@ -1960,10 +2062,6 @@ void glrender(const string& prefix, const picture *pic, const string& format,
 #ifdef HAVE_LIBGLUT
     setosize();
 #endif
-
-    if(View && settings::verbose > 1)
-      cout << "Rendering " << stripDir(prefix) << " as "
-           << Width << "x" << Height << " image" << endl;
 #endif
   }
 
@@ -2066,13 +2164,21 @@ void glrender(const string& prefix, const picture *pic, const string& format,
 
   initialized=true;
 
+#if defined(HAVE_COMPUTE_SHADER) && !defined(HAVE_LIBOSMESA)
+  GPUindexing=getSetting<bool>("GPUindexing");
+  GPUcompress=getSetting<bool>("GPUcompress");
+#else
+  GPUindexing=false;
+  GPUcompress=false;
+#endif
+
   GLint val;
   glGetIntegerv(GL_MAX_UNIFORM_BLOCK_SIZE,&val);
 
   if(GPUindexing) {
-    glGetIntegerv(GL_MAX_COMPUTE_WORK_GROUP_INVOCATIONS,&processors);
-    if(processors <= 1)
-      GPUindexing=false;
+    gl::localSize=getSetting<Int>("GPUlocalSize");
+    gl::blockSize=getSetting<Int>("GPUblockSize");
+    gl::groupSize=gl::localSize*gl::blockSize;
   }
 
   Maxmaterials=val/sizeof(Material);
@@ -2216,71 +2322,160 @@ void registerBuffer(const std::vector<T>& buffervector, GLuint& bufferIndex,
 void clearCount()
 {
   glUseProgram(zeroShader);
-  glUniform1ui(glGetUniformLocation(zeroShader,"width"),gl::Width);
   gl::lastshader=zeroShader;
+  glUniform1ui(glGetUniformLocation(zeroShader,"width"),gl::Width);
   fpu_trap(false); // Work around FE_INVALID
   glDrawArrays(GL_TRIANGLES, 0, 3);
   fpu_trap(settings::trap());
   glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
 }
 
+void compressCount()
+{
+  glMemoryBarrier(GL_BUFFER_UPDATE_BARRIER_BIT);
+  glUseProgram(compressShader);
+  gl::lastshader=compressShader;
+  glUniform1ui(glGetUniformLocation(compressShader,"width"),gl::Width);
+  fpu_trap(false); // Work around FE_INVALID
+  glDrawArrays(GL_TRIANGLES, 0, 3);
+  fpu_trap(settings::trap());
+  glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
+}
+
+void partialSums(bool readSize=false)
+{
+  // Compute partial sums on the GPU
+  glUseProgram(sum1Shader);
+  glDispatchCompute(gl::g,1,1);
+
+  if(gl::elements <= gl::groupSize*gl::groupSize)
+    glUseProgram(sum2fastShader);
+  else {
+    glUseProgram(sum2Shader);
+    glUniform1ui(glGetUniformLocation(sum2Shader,"blockSize"),
+                 gl::ceilquotient(gl::g,gl::localSize));
+  }
+  glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
+  glDispatchCompute(1,1,1);
+
+  glUseProgram(sum3Shader);
+  glUniform1ui(glGetUniformLocation(sum3Shader,"final"),gl::elements-1);
+  glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
+  glDispatchCompute(gl::g,1,1);
+}
+
+void resizeFragmentBuffer()
+{
+  if(GPUindexing) {
+    glMemoryBarrier(GL_BUFFER_UPDATE_BARRIER_BIT);
+
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::feedbackBuffer);
+    GLuint *feedback=(GLuint *) glMapBuffer(GL_SHADER_STORAGE_BUFFER,GL_READ_ONLY);
+
+    GLuint maxDepth=feedback[0];
+    if(maxDepth > gl::maxSize)
+      gl::resizeBlendShader(maxDepth);
+
+    fragments=feedback[1];
+    glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);
+  }
+
+  if(fragments > maxFragments) {
+    // Initialize the alpha buffer
+    maxFragments=11*fragments/10;
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::fragmentBuffer);
+    glBufferData(GL_SHADER_STORAGE_BUFFER,maxFragments*sizeof(glm::vec4),
+                 NULL,GL_DYNAMIC_DRAW);
+    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,4,camp::fragmentBuffer);
+
+
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::depthBuffer);
+    glBufferData(GL_SHADER_STORAGE_BUFFER,maxFragments*sizeof(GLfloat),
+                 NULL,GL_DYNAMIC_DRAW);
+    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,5,camp::depthBuffer);
+
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::feedbackBuffer);
+  }
+}
+
 void refreshBuffers()
 {
   GLuint zero=0;
-  GLuint fragments=0;
-  GLuint pixels=gl::Width*gl::Height;
+  gl::pixels=(gl::Width+1)*(gl::Height+1);
 
   if(initSSBO) {
+    gl::processors=1;
+
+    GLuint Pixels;
+    if(GPUindexing) {
+      GLuint G=gl::ceilquotient(gl::pixels,gl::groupSize);
+      Pixels=gl::groupSize*G;
+
+      GLuint globalSize=gl::localSize*gl::ceilquotient(G,gl::localSize);
+      glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::globalSumBuffer);
+      glBufferData(GL_SHADER_STORAGE_BUFFER,globalSize*sizeof(GLuint),NULL,
+                   GL_DYNAMIC_READ);
+      glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R32UI,GL_RED_INTEGER,
+                        GL_UNSIGNED_INT,&zero);
+      glBindBufferBase(GL_SHADER_STORAGE_BUFFER,3,camp::globalSumBuffer);
+    } else Pixels=gl::pixels;
+
     glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::offsetBuffer);
-    glBufferData(GL_SHADER_STORAGE_BUFFER,(pixels+GPUindexing)*sizeof(GLuint),
+    glBufferData(GL_SHADER_STORAGE_BUFFER,(Pixels+2)*sizeof(GLuint),
                  NULL,GL_DYNAMIC_DRAW);
-    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,camp::offsetBuffer);
-    glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R8UI,GL_RED_INTEGER,
-                      GL_UNSIGNED_BYTE,&zero);
+    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,camp::offsetBuffer);
 
     glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::countBuffer);
-    glBufferData(GL_SHADER_STORAGE_BUFFER,pixels*sizeof(GLuint),NULL,
-                 GL_DYNAMIC_DRAW);
+    glBufferData(GL_SHADER_STORAGE_BUFFER,(Pixels+2)*sizeof(GLuint),
+                 NULL,GL_DYNAMIC_DRAW);
     glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,camp::countBuffer);
-    glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R8UI,GL_RED_INTEGER,
-                      GL_UNSIGNED_BYTE,&zero);
+
+    if(GPUcompress) {
+      glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::indexBuffer);
+      glBufferData(GL_SHADER_STORAGE_BUFFER,gl::pixels*sizeof(GLuint),
+                   NULL,GL_DYNAMIC_DRAW);
+      glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,camp::indexBuffer);
+
+      GLuint one=1;
+      glBindBuffer(GL_ATOMIC_COUNTER_BUFFER,camp::elementsBuffer);
+      glBufferData(GL_ATOMIC_COUNTER_BUFFER,sizeof(GLuint),&one,
+                   GL_DYNAMIC_DRAW);
+      glBindBufferBase(GL_ATOMIC_COUNTER_BUFFER,0,camp::elementsBuffer);
+    } else
+      glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R8UI,GL_RED_INTEGER,
+                        GL_UNSIGNED_BYTE,&zero); // Clear the count buffer
 
     glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::opaqueBuffer);
-    glBufferData(GL_SHADER_STORAGE_BUFFER,pixels*sizeof(glm::vec4),NULL,GL_DYNAMIC_DRAW);
-    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,5,camp::opaqueBuffer);
+    glBufferData(GL_SHADER_STORAGE_BUFFER,gl::pixels*sizeof(glm::vec4),NULL,
+                 GL_DYNAMIC_DRAW);
+    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,6,camp::opaqueBuffer);
 
     glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::opaqueDepthBuffer);
-    glBufferData(GL_SHADER_STORAGE_BUFFER,pixels*sizeof(GLfloat),NULL,GL_DYNAMIC_DRAW);
-    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,6,camp::opaqueDepthBuffer);
+    glBufferData(GL_SHADER_STORAGE_BUFFER,
+                 sizeof(GLuint)+gl::pixels*sizeof(GLfloat),NULL,
+                 GL_DYNAMIC_DRAW);
+    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,7,camp::opaqueDepthBuffer);
     const GLfloat zerof=0.0;
     glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R32F,GL_RED,GL_FLOAT,&zerof);
 
-    if(GPUindexing) {
-      glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::sumBuffer);
-      glBufferData(GL_SHADER_STORAGE_BUFFER,(gl::processors+1)*sizeof(GLuint),NULL,
-                   GL_DYNAMIC_DRAW);
-      glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R8UI,GL_RED_INTEGER,
-                        GL_UNSIGNED_BYTE,&zero);
-      glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,camp::sumBuffer);
-    }
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::feedbackBuffer);
+    glBufferData(GL_SHADER_STORAGE_BUFFER,2*sizeof(GLuint),NULL,
+                 GL_DYNAMIC_DRAW);
+    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,8,camp::feedbackBuffer);
 
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::feedbackBuffer);
     initSSBO=false;
   }
 
-  if(GPUindexing && gl::exporting) {
+  // Determine the fragment offsets
+
+  if(gl::exporting && GPUindexing && !GPUcompress) {
     glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::countBuffer);
     glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R8UI,GL_RED_INTEGER,
                       GL_UNSIGNED_BYTE,&zero);
-
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::offsetBuffer);
-    glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R8UI,GL_RED_INTEGER,
-                      GL_UNSIGNED_BYTE,&zero);
-
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::sumBuffer);
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::feedbackBuffer);
   }
 
-  // Determine the fragment offsets
-
   if(!interlock) {
     drawBuffer(material1Data,countShader);
     drawBuffer(materialData,countShader);
@@ -2294,32 +2489,58 @@ void refreshBuffers()
   glEnable(GL_MULTISAMPLE);
   glDepthMask(GL_TRUE); // Write to depth buffer
 
-  if(GPUindexing) { // Compute partial sums directly on the GPU
-    glUseProgram(preSumShader);
-    glUniform1ui(glGetUniformLocation(preSumShader,"elements"),pixels);
-    glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
-    glDispatchCompute(gl::processors,1,1);
+  if(GPUcompress) {
+    compressCount();
+    GLuint *p=(GLuint *) glMapBuffer(GL_ATOMIC_COUNTER_BUFFER,GL_READ_WRITE);
+    gl::elements=GPUindexing ? p[0] : p[0]-1;
+    p[0]=1;
+    glUnmapBuffer(GL_ATOMIC_COUNTER_BUFFER);
+    if(gl::elements == 0) return;
+  } else
+    gl::elements=gl::pixels;
 
-    glUseProgram(partialSumShader);
-    glUniform1ui(glGetUniformLocation(partialSumShader,"elements"),pixels);
+  if(GPUindexing) {
+    gl::g=gl::ceilquotient(gl::elements,gl::groupSize);
+    gl::elements=gl::groupSize*gl::g;
 
-    glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
-    glDispatchCompute(1,1,1);
+    if(settings::verbose > 3) {
+      static bool first=true;
+      if(first) {
+        partialSums();
+        first=false;
+      }
+      unsigned int N=10000;
+      seconds();
+      for(unsigned int i=0; i < N; ++i)
+        partialSums();
+      glFinish();
+      double T=seconds()/N;
+      cout << "elements=" << gl::elements << endl;
+      cout << "Tmin (ms)=" << T*1e3 << endl;
+      cout << "Megapixels/second=" << gl::elements/T/1e6 << endl;
+    }
 
-    glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
-    glGetBufferSubData(GL_SHADER_STORAGE_BUFFER,gl::processors*sizeof(GLuint),sizeof(GLuint),&fragments);
-  } else { // Compute partial sums on the CPU
+    partialSums(true);
+  } else {
+    size_t size=gl::elements*sizeof(GLuint);
+
+    // Compute partial sums on the CPU
     glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::countBuffer);
-    GLuint *countm1=(GLuint *) (glMapBuffer(GL_SHADER_STORAGE_BUFFER,GL_READ_ONLY))-1;
+    GLuint *p=(GLuint *) glMapBufferRange(GL_SHADER_STORAGE_BUFFER,
+                                          0,size+sizeof(GLuint),
+                                              GL_MAP_READ_BIT);
+    GLuint maxsize=p[0];
+    GLuint *count=p+1;
 
     glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::offsetBuffer);
-    GLuint *offset=(GLuint *) glMapBuffer(GL_SHADER_STORAGE_BUFFER,GL_WRITE_ONLY);
+    GLuint *offset=(GLuint *) glMapBufferRange(GL_SHADER_STORAGE_BUFFER,
+                                               sizeof(GLuint),size,
+                                               GL_MAP_WRITE_BIT);
 
-    size_t Offset=0;
-    offset[0]=0;
-    for(size_t i=1; i < pixels; ++i)
-      offset[i]=Offset += countm1[i];
-    fragments=offset[pixels-1]+countm1[pixels];
+    size_t Offset=offset[0]=count[0];
+    for(size_t i=1; i < gl::elements; ++i)
+      offset[i]=Offset += count[i];
+    fragments=Offset;
 
     glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::offsetBuffer);
     glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);
@@ -2327,26 +2548,16 @@ void refreshBuffers()
     glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::countBuffer);
     glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);
 
-    clearCount();
+    if(gl::exporting) {
+      glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::countBuffer);
+      glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R32UI,GL_RED_INTEGER,
+                        GL_UNSIGNED_INT,&zero);
+    } else
+      clearCount();
+
+    if(maxsize > gl::maxSize)
+      gl::resizeBlendShader(maxsize);
   }
-
-  if(fragments > maxFragments) {
-    // Initialize the alpha buffer
-    maxFragments=11*fragments/10;
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::fragmentBuffer);
-    glBufferData(GL_SHADER_STORAGE_BUFFER,maxFragments*sizeof(glm::vec4),
-                 NULL,GL_DYNAMIC_DRAW);
-    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,3,camp::fragmentBuffer);
-
-
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::depthBuffer);
-    glBufferData(GL_SHADER_STORAGE_BUFFER,maxFragments*sizeof(GLfloat),
-                 NULL,GL_DYNAMIC_DRAW);
-    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,4,camp::depthBuffer);
-
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::sumBuffer);
-  }
-
   gl::lastshader=-1;
 }
 
@@ -2356,16 +2567,27 @@ void setUniforms(vertexBuffer& data, GLint shader)
 
   if(shader != gl::lastshader) {
     glUseProgram(shader);
-    gl::lastshader=shader;
 
-    if(camp::ssbo) {
+    if(normal)
       glUniform1ui(glGetUniformLocation(shader,"width"),gl::Width);
-      GLuint pixels=gl::Width*gl::Height;
-      GLuint M=GPUindexing ? pixels/gl::processors : 0;
-      GLuint r=pixels-M*gl::processors;
-      glUniform1ui(glGetUniformLocation(shader,"M"),M);
-      glUniform1ui(glGetUniformLocation(shader,"r"),r);
-    }
+  }
+
+  glUniformMatrix4fv(glGetUniformLocation(shader,"projViewMat"),1,GL_FALSE,
+                     value_ptr(gl::projViewMat));
+
+  glUniformMatrix4fv(glGetUniformLocation(shader,"viewMat"),1,GL_FALSE,
+                     value_ptr(gl::viewMat));
+  if(normal)
+    glUniformMatrix3fv(glGetUniformLocation(shader,"normMat"),1,GL_FALSE,
+                       value_ptr(gl::normMat));
+
+  if(shader == countShader) {
+    gl::lastshader=shader;
+    return;
+  }
+
+  if(shader != gl::lastshader) {
+    gl::lastshader=shader;
     glUniform1ui(glGetUniformLocation(shader,"nlights"),gl::nlights);
 
     for(size_t i=0; i < gl::nlights; ++i) {
@@ -2382,10 +2604,13 @@ void setUniforms(vertexBuffer& data, GLint shader)
                   (GLfloat) gl::Diffuse[i4+2]);
     }
 
-    if (settings::getSetting<bool>("ibl")) {
-      gl::IBLbrdfTex.setUniform(glGetUniformLocation(shader,"reflBRDFSampler"));
-      gl::irradiance.setUniform(glGetUniformLocation(shader,"diffuseSampler"));
-      gl::reflTextures.setUniform(glGetUniformLocation(shader,"reflImgSampler"));
+    if(settings::getSetting<bool>("ibl")) {
+      gl::IBLbrdfTex.setUniform(glGetUniformLocation(shader,
+                                                     "reflBRDFSampler"));
+      gl::irradiance.setUniform(glGetUniformLocation(shader,
+                                                     "diffuseSampler"));
+      gl::reflTextures.setUniform(glGetUniformLocation(shader,
+                                                       "reflImgSampler"));
     }
   }
 
@@ -2395,15 +2620,6 @@ void setUniforms(vertexBuffer& data, GLint shader)
   bool copy=(gl::remesh || data.partial || !data.rendered) && !gl::copied;
   registerBuffer(data.materials,data.materialsBuffer,copy,GL_UNIFORM_BUFFER);
   glBindBufferBase(GL_UNIFORM_BUFFER,binding,data.materialsBuffer);
-
-  glUniformMatrix4fv(glGetUniformLocation(shader,"projViewMat"),1,GL_FALSE,
-                     value_ptr(gl::projViewMat));
-
-  glUniformMatrix4fv(glGetUniformLocation(shader,"viewMat"),1,GL_FALSE,
-                     value_ptr(gl::viewMat));
-  if(normal)
-    glUniformMatrix3fv(glGetUniformLocation(shader,"normMat"),1,GL_FALSE,
-                       value_ptr(gl::normMat));
 }
 
 void drawBuffer(vertexBuffer& data, GLint shader, bool color)
@@ -2510,16 +2726,11 @@ void aBufferTransparency()
   // Blend transparent fragments
   glDisable(GL_DEPTH_TEST);
   glUseProgram(blendShader);
+  gl::lastshader=blendShader;
   glUniform1ui(glGetUniformLocation(blendShader,"width"),gl::Width);
-  GLuint pixels=gl::Width*gl::Height;
-  GLuint M=GPUindexing ? pixels/gl::processors : 0;
-  GLuint r=pixels-M*gl::processors;
-  glUniform1ui(glGetUniformLocation(blendShader,"M"),M);
-  glUniform1ui(glGetUniformLocation(blendShader,"r"),r);
   glUniform4f(glGetUniformLocation(blendShader,"background"),
               gl::Background[0],gl::Background[1],gl::Background[2],
               gl::Background[3]);
-  gl::lastshader=blendShader;
   fpu_trap(false); // Work around FE_INVALID
   glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
   glDrawArrays(GL_TRIANGLES,0,3);
@@ -2552,7 +2763,10 @@ void drawBuffers()
   if(camp::ssbo) {
     if(transparent) {
       refreshBuffers();
-      if(!interlock) gl::copied=true;
+      if(!interlock) {
+        resizeFragmentBuffer();
+        gl::copied=true;
+      }
     }
   }
 
@@ -2564,6 +2778,7 @@ void drawBuffers()
 
   if(transparent) {
     gl::copied=true;
+    if(interlock) resizeFragmentBuffer();
     drawTransparent();
   }
   Opaque=0;
