@@ -71,7 +71,6 @@ using vm::array;
 
 void runFile(const string& filename);
 
-
 namespace settings {
 
 using camp::pair;
@@ -950,6 +949,7 @@ struct versionOption : public option {
     bool ssbo=false;
     bool gsl=false;
     bool fftw3=false;
+    bool eigen=false;
     bool xdr=false;
     bool curl=false;
     bool lsp=false;
@@ -979,6 +979,10 @@ struct versionOption : public option {
     fftw3=true;
 #endif
 
+#ifdef HAVE_EIGEN_DENSE
+    eigen=true;
+#endif
+
 #ifdef HAVE_RPC_RPC_H
     xdr=true;
 #endif
@@ -1023,6 +1027,7 @@ struct versionOption : public option {
     feature("SSBO     GLSL shader storage buffer objects",ssbo);
     feature("GSL      GNU Scientific Library (special functions)",gsl);
     feature("FFTW3    Fast Fourier transforms",fftw3);
+    feature("Eigen    Eigenvalue library",eigen);
     feature("XDR      External Data Representation (portable binary file format for V3D)",xdr);
     feature("CURL     URL support",curl);
     feature("LSP      Language Server Protocol",lsp);
@@ -1248,6 +1253,13 @@ void initSettings() {
                             "Compute indexing partial sums on GPU", true));
   addOption(new boolSetting("GPUinterlock", 0,
                             "Use fragment shader interlock", true));
+  addOption(new boolSetting("GPUcompress", 0,
+                            "Compress GPU transparent fragment counts",
+                            false));
+  addOption(new IntSetting("GPUlocalSize", 0, "n",
+                           "Compute shader local size", 256));
+  addOption(new IntSetting("GPUblockSize", 0, "n",
+                           "Compute shader block size", 8));
 
   addOption(new pairSetting("position", 0, "pair",
                             "Initial 3D rendering screen position"));
@@ -1450,7 +1462,7 @@ void initSettings() {
   addOption(new stringSetting("dvipsOptions", 0, "string", ""));
   addOption(new stringSetting("dvisvgmOptions", 0, "string", ""));
   addOption(new boolSetting("dvisvgmMultipleFiles", 0,
-                            "dvisvgm supports multiple files", false));
+                            "dvisvgm supports multiple files", true));
   addOption(new stringSetting("convertOptions", 0, "string", ""));
   addOption(new stringSetting("gsOptions", 0, "string", ""));
   addOption(new stringSetting("htmlviewerOptions", 0, "string", ""));
